@@ -48,16 +48,16 @@ class UserAuthorize {
 
     // 桼ѹ
     if (isset($_GET['brows_name'])) {
-      // !!! ¸ߤ桼ǧ٤
-      $_SESSION['brows_name'] = $_GET['brows_name'];
+      // ¸ߤ桼ǧ
+      if (in_array($_GET['brows_name'], $this->getUserNames())) {
+	$_SESSION['brows_name'] = $_GET['brows_name'];
+      }
     }
   }
 
 
   /*!
     \brief 
-
-    \todo ǧڽ
   */
   function login($name, $password) {
 
@@ -70,6 +70,7 @@ class UserAuthorize {
     $this->login_fail = true;
     $fp = @fopen($_SESSION['users_file'], 'r');
     if ($fp != false) {
+      flock($fp, LOCK_SH);
       while (! feof($fp)) {
 	$line = fgets($fp);
 	$user = strtok($line, "\n\t ");
@@ -88,6 +89,7 @@ class UserAuthorize {
       // 桼ե뤬¸ߤʤ硢admin ѥɤɲä
       $fp = fopen($_SESSION['users_file'], 'w');
       if ($fp != false) {
+	flock($fp, LOCK_EX);
 	$line = "admin\t". crypt('admin'). "\n";
 	fwrite($fp, $line);
 	fclose($fp);
@@ -119,7 +121,7 @@ class UserAuthorize {
     \brief 󤷤Ƥп֤
   */
   function isLogin() {
-    return (array_key_exists('login_name', $_SESSION)) ? true : false;
+    return isset($_SESSION['login_name']) ? true : false;
   }
 
 
@@ -163,6 +165,7 @@ class UserAuthorize {
     $users = array();
     $fp = @fopen($_SESSION['users_file'], 'r');
     if ($fp != false) {
+      flock($fp, LOCK_SH);
       while (! feof($fp)) {
 	$line = fgets($fp);
 	$user = strtok($line, "\n\t ");
@@ -189,6 +192,16 @@ class LangUpload {
   var $comments_mtime;
   var $send_comment;
   var $program_updated;
+  var $rev_max;			//!< ӥֹκ
+  var $rev;			//!< ӥֹ
+  var $key;			//!< ɽؼ
+  var $read_mtime;		//!< ɻ
+
+  // !!! ɬפ˱ơѿ
+  // !!! ¿ʬ
+  var $files_mtime;		//!< եι
+
+  var $unread_pages;		//!< ̤ɥե
 
 
   /*!
@@ -200,7 +213,7 @@ class LangUpload {
   function LangUpload() {
 
     // ե뤫νͤɤ߽Ф
-    if (! array_key_exists('admin_mail_address', $_SESSION)) {
+    if (! isset($_SESSION['admin_mail_address'])) {
       require_once('config.php');
 
       $_SESSION['admin_mail_address'] =
@@ -221,6 +234,46 @@ class LangUpload {
     $this->comments_mtime = array();
     $this->send_comment = false;
     $this->program_updated = false;
+    $this->read_mtime = array();
+    $this->files_mtime = array();
+    $this->unread_pages = array();
+
+    // եι
+    if (isset($_SESSION['brows_name'])) {
+      $this->createFilesHash($_SESSION['brows_name']);
+    }
+
+    // ڡѹ
+    if (isset($_GET['chapter'])) {
+      // ¸ߤϤǧ
+      $chapter = intval($_GET['chapter']);
+      if (($chapter > 0) && ($chapter <= count($_SESSION['problems_num']))) {
+	$_SESSION['chapter'] = $chapter;
+      }
+    }
+    if (isset($_GET['no'])) {
+      // ¸ߤֹ椫ǧ
+      $no = intval($_GET['no']);
+      if (($no > 0) &&
+	  ($no <= $_SESSION['problems_num'][$_SESSION['chapter'] -1])) {
+	$_SESSION['no'] = intval($_GET['no']);
+      }
+    }
+
+    // key η׻
+    $this->key = (isset($_SESSION['chapter']) && isset($_SESSION['no'])) ?
+      $_SESSION['chapter']. '-'. $_SESSION['no'] : '0-0';
+
+    // ɽΥӥֹ
+    $this->rev_max = isset($this->programs_exist[$this->key]) ?
+      $this->programs_exist[$this->key] : 0;
+    $this->rev = $this->rev_max;
+    if (isset($_GET['rev'])) {
+      $rev = intval($_GET['rev']);
+      if (($rev >= 0) && ($rev <= $this->rev_max)) {
+	$this->rev = $rev;
+      }
+    }
 
     // åץɥեν
     if (isset($_FILES) && isset($_FILES['userfile']['tmp_name'])) {
@@ -228,15 +281,8 @@ class LangUpload {
 			   $_FILES['userfile']['name']);
     }
 
-    // ڡѹ
-    if (isset($_GET['chapter'])) {
-      // !!! ¸ߤϤǧ٤
-      $_SESSION['chapter'] = intval($_GET['chapter']);
-    }
-    if (isset($_GET['no'])) {
-      // !!! ¸ߤֹ椫ǧ٤
-      $_SESSION['no'] = intval($_GET['no']);
-    }
+    // ɤδ򹹿
+    $this->selectUnreadPage();
   }
 
 
@@ -262,9 +308,10 @@ class LangUpload {
     if ($fp == false) {
       $out .= 'no file: version.txt';
     } else {
+      flock($fp, LOCK_SH);
       while (! feof($fp)) {
 	$line = fgets($fp);
-	if (preg_match('/(\d\d\d\d\/\d\d\/\d\d)/', $line, $matches)) {
+	if (preg_match('/(\d\d\d\d\/\d\d\/\d\d)/', $line)) {
 	  $out .= '<h3>'. $line. '</h3>';
 	} else if (preg_match('/---/', $line)) {
 	  continue;
@@ -298,24 +345,20 @@ class LangUpload {
     \brief եΥåץɽ
   */
   function uploadProgram($uploaded_file, $file_name) {
-    $login_name = $_SESSION['login_name'];
-    $this->createFilesHash($login_name);
 
-    $uploaddir = 'data/'. $login_name. '/';
-    $key = $this->getChapter(). '-'. $this->getNo();
-    $rev = array_key_exists($key, $this->programs_exist) ?
-      $this->programs_exist[$key] : 0;
-    $next_rev = $rev + 1;
+    $next_rev = $this->rev + 1;
+    $file = 'p'. $this->key. '-'. $next_rev. '.txt';
 
-    $file = 'p'. $key. '-'. $next_rev. '.txt';
+    $uploaddir = 'data/'. $_SESSION['login_name']. '/';
     $to_file = $uploaddir. $file;
     if (! move_uploaded_file($uploaded_file, $to_file)) {
       // print "Possible file upload attack!\n";
       // !!! ȤΤ˻Ĥ
     } else {
-      // !!! file_name 򸵤ˡprograms.txt 
+      // åץɻΥե̾Ͽ
       $fp = fopen($uploaddir. 'programs.txt', 'a');
       if ($fp != false) {
+	flock($fp, LOCK_EX);
 	$line = $file. "\t". $file_name. "\n";
 	fwrite($fp, $line);
 	fclose($fp);
@@ -357,37 +400,21 @@ class LangUpload {
   }
 
 
-  /*!
-    \brief ϥ󥯤κ
-
-    \todo ̤ɥȤΤֹϡ
-  */
-  function createChapterLink($chapter_label = 'Chap.') {
-    $out = '';
-    $out .= '<table><tr><th>'. $chapter_label. '</th>';
-
-    $index = 1;
-    foreach ($_SESSION['problems_num'] as $chapter) {
-      $out .= '<td><a href=?chapter='.
-	$index. '>&nbsp;'. $index. '&nbsp;</a></td>';
-      ++$index;
-    }
-    $out .= '</tr></table>';
-    return $out;
-  }
-
-
   /*!
     \brief ֹ󥯤κ
 
     \todo ̤ɥȤΤֹϡ
     \todo ץब¸ߤСʬ褦ˤ
   */
-  function createNoLink($chapter, $no_label = 'No.') {
+  function createNoLink($chapter, $no_label = '%d-%d') {
     $out = '';
-    $out .= '<table><tr><th>'. $no_label. '</th>';
 
     $n = $_SESSION['problems_num'][$chapter -1];
+    $no = $this->getNo();
+    $chapter = $this->getChapter();
+    $no_title = sprintf($no_label, $chapter, $no);
+    $out .= '<table border=1><tr><th colspan='. $n. ' align=left>&nbsp;'. $this->createPrevNoLink(). '&nbsp;'. $no_title. '&nbsp;'. $this->createNextNoLink(). '</th></tr><tr>';
+
     for ($i = 0; $i < $n; ++$i) {
       $index = $i + 1;
       $out .= '<td><a href=?no='. $index. '>&nbsp;'. $index. '</a></td>';
@@ -397,20 +424,11 @@ class LangUpload {
   }
 
 
-  /*!
-    \brief ̤ɥ󥯤ɤ
-  */
-  function createUnreadLink() {
-    $out = '';
-    $out .= '!!! ̤ɥȤɤ';
-
-    return $out;
-  }
-
-
   /*!
     \brief ᥤڡѤΥơ֥
 
+    \todo Ȥͽ꤬ʤС
+
     \todo Smarty ΰ̣礤̵...Ĥ
     \todo ̤ɥ󥯤ξϤʬ褦ɽ롣Ĥơ֥Ǥ狼뤫...
   */
@@ -447,9 +465,6 @@ class LangUpload {
       $total_exists += $exists_num;
       $total_problems += $total_num;
 
-      // Ȥ̵ͭΥղ
-      // !!! ̤
-
       $ret .= '</td></tr>';
       ++$chapter_no;
     }
@@ -478,40 +493,44 @@ class LangUpload {
   function createProgramsTable($brows_name) {
 
     $chapter = $this->getChapter();
-    if (! array_key_exists($chapter - 1, $_SESSION['problems_num'])) {
+    if (! isset($_SESSION['problems_num'][$chapter - 1])) {
+      // !!! ꤷֹ꤬ʤ
       // !!! åϤ٤
+      // !!! ˽񤭹٤
       return;
     }
-    $this->createFilesHash($brows_name);
 
     $out = '';
     $out .= '<table border=1>';
-    $out .= '<tr><th>No.</th><th>Program</th><th>Comment</th></tr>';
+    $out .= '<tr><th>No.</th><th width=110>Program</th><th width=110>Comment</th></tr>';
 
     for ($i = 0; $i < $_SESSION['problems_num'][$chapter - 1]; ++$i) {
       $no = $i + 1;
-      $out .= '<tr><td><a href=programView.php?chapter='. $chapter.
-	'&no='. $no. '>'. $chapter. '-'. $no. '</a></td>';
+      $ref_front = '<a href=programView.php?chapter='. $chapter. '&no='. $no;
+      $out .= '<tr><td align=center>'. $ref_front. '>&nbsp;'. $chapter. '-'. $no. '&nbsp;</a></td>';
       $key = $chapter. '-'. ($i + 1);
 
       // ץ¸ɽ
-      $out .= '<td>&nbsp;';
-      if (array_key_exists($key, $this->programs_exist)) {
+      $out .= '<td align=center>';
+      if (isset($this->programs_exist[$key])) {
 	$rev = $this->programs_exist[$key];
 	$mtime = $this->programs_mtime[$key. '-'. $rev];
-	$out .= date('Y-m-d', $mtime);
+	$out .= $ref_front. '>&nbsp;'. date('Y-m-d', $mtime). '&nbsp;</a>';
+      } else {
+	$out .= '&nbsp;';
       }
-      $out .= '&nbsp;</td>';
+      $out .= '</td>';
 
       // Ȥ¸ɽ
-      $out .= '<td>&nbsp;';
-      if (array_key_exists($key, $this->comments_exist)) {
+      $out .= '<td align=center>';
+      if (isset($this->comments_exist[$key])) {
 	$rev = $this->comments_exist[$key];
 	$mtime = $this->comments_mtime[$key. '-'. $rev];
-	$out .= date('Y-m-d', $mtime);
+	$out .= $ref_front. '#comment>&nbsp;'. date('Y-m-d', $mtime). '&nbsp;</a>';
+      } else {
+	$out .= '&nbsp;';
       }
-      $out .= '&nbsp;</td>';
-      $out .= '</tr>';
+      $out .= '</td></tr>';
     }
     $out .= '</table>';
     return $out;
@@ -528,28 +547,25 @@ class LangUpload {
   function createProgramView($brows_name,
 			     $no_program_message = 'No Program.') {
 
-    $chapter = $this->getChapter();
-    $no = $this->getNo();
-    $this->createFilesHash($brows_name);
-
     $out = '';
-
-    $key = $chapter. '-'. $no;
-    if (! array_key_exists($key, $this->programs_exist)) {
-      // !!! ե뤬ʤν
-      // !!! ɽʤ餤ʡ
+    if (! isset($this->programs_exist[$this->key])) {
+      // ե뤬ʤν
       $out .= $no_program_message;
       return $out;
     }
-    $rev = $this->programs_exist[$key];
-    $path = 'data/'. $brows_name. '/p'. $key. '-'. $rev. '.txt';
+    $rev = $this->rev;
+    $path = 'data/'. $brows_name. '/p'. $this->key. '-'. $rev. '.txt';
 
+    // !!! enscript ʤȤե⡼ɤΤȤб
     $cmd = $_SESSION['enscript_bin'].
       ' --highlight=cpp --color --language html --toc --output - '. $path;
-    ereg('(<PRE>.+</PRE>)', shell_exec($cmd), $prog);
+    $stdout = shell_exec($cmd);
+    $code = mb_detect_encoding($stdout, "auto", true);
+    $converted = ($code != false) ?
+      mb_convert_encoding($stdout, 'EUC-JP', $code) : $stdout;
+    ereg('(<PRE>.+</PRE>)', $converted, $prog);
 
     $out .= $prog[1];
-    // !!! ʤˤʤСλݤ֤
     return $out;
   }
 
@@ -583,17 +599,17 @@ class LangUpload {
       return;
     }
 
-    clearstatcache();		// !!! ɬפ
+    clearstatcache();
     while (($file = readdir($dp)) !== false) {
-      if (preg_match('/p([\d]+-[\d])+-([\d])+\.txt/', $file, $matches)) {
+      if (preg_match('/p([\d]+-[\d]+)-([\d]+)\.txt/', $file, $matches)) {
 	// pX-X-X.txt Ȥե뤬п
 	$key = $matches[1];
-	$rev = $matches[2];
+	$rev = intval($matches[2]);
 	$this->programs_mtime[$key. '-'. $rev] =
 	  filemtime($dir_name. '/'. $file);
 
 	// ץ
-	if (array_key_exists($key, $this->programs_exist)) {
+	if (isset($this->programs_exist[$key])) {
 	  $current_rev = $this->programs_exist[$key];
 	  if ($rev > $current_rev) {
 	    $this->programs_exist[$key] = $rev;
@@ -603,15 +619,15 @@ class LangUpload {
 	}
       }
 
-      if (preg_match('/c([\d]+-[\d])+-([\d])+\.txt/', $file, $matches)) {
+      if (preg_match('/c([\d]+-[\d]+)-([\d]+)\.txt/', $file, $matches)) {
 	// cX-X-X.txt Ȥե뤬п
 	$key = $matches[1];
-	$rev = $matches[2];
+	$rev = intval($matches[2]);
 	$this->comments_mtime[$key. '-'. $rev] =
 	  filemtime($dir_name. '/'. $file);
 
 	// ȴ
-	if (array_key_exists($key, $this->comments_exist)) {
+	if (isset($this->comments_exist[$key])) {
 	  $current_rev = $this->comments_exist[$key];
 	  if ($rev > $current_rev) {
 	    $this->comments_exist[$key] = $rev;
@@ -621,9 +637,7 @@ class LangUpload {
 	}
       }
     }
-
-    $ret = array();
-    return $ret;
+    closedir($dp);
   }
 
 
@@ -633,7 +647,6 @@ class LangUpload {
     \todo Ƥȴؿ̾פƤʤΤ
   */
   function getProgramExists($brows_name) {
-    $this->createFilesHash($brows_name);
 
     $exists = array();
     $index = 1;
@@ -641,7 +654,7 @@ class LangUpload {
       $chapter_programs = 0;
       for ($i = 1; $i <= $chapter_problems; ++$i) {
 	$key = $index. '-'. $i;
-	if (array_key_exists($key, $this->programs_exist)) {
+	if (isset($this->programs_exist[$key])) {
 	  ++$chapter_programs;
 	}
       }
@@ -656,13 +669,13 @@ class LangUpload {
     \brief ưѤΥ󥯤
   */
   function createInternalLink($chapter, $no) {
-    $out = '<a href=main.php>Main Page</a>';
+    $out = '<a href=main.php><strong>Main Page</strong></a>';
     if (($chapter > 0) && ($no <= 0)) {
       $out .= ' &gt;&gt; <strong>Chapter Page</strong>';
     }
     if (($chapter > 0) && ($no > 0)) {
       $out .= ' &gt;&gt; <a href=chapterView.php?chapter='. $chapter.
-	'>Chapter Page</a> &gt;&gt; <strong>Program Page</strong>';
+	'><strong>Chapter Page</strong></a> &gt;&gt; <strong>Program Page</strong>';
     }
     return $out;
   }
@@ -678,15 +691,26 @@ class LangUpload {
     $users = $user->getUserNames();
 
     $out = '';
+    $out .= '<table class=users width=100%><tr><td>&nbsp;</td>';
     foreach ($users as $each_name) {
       if ($_SESSION['brows_name'] == $each_name) {
-	// !!! Ĵɽ
-	$out .= $each_name;
+	$out .= '<td class=users><img src=front_left.png></td><td background=front_middle.png>'. $each_name.
+	  '</td><td class=users><img src=front_right.png></td>';
       } else {
-	$out .= '<a href=?brows_name='. $each_name. '>'. $each_name. '</a>';
+	$out .= '<td class=users><img src=back_left.png></td><td background=back_middle.png><a class=users href=?brows_name='.
+	  $each_name. '>'. $each_name. '</a></td><td class=users><img src=back_right.png></td>';
       }
-      $out .= ' ';
     }
+    $out .= '<td width=100%></td></tr><tr><td><img src=dot.png width=100% height=2</td>';
+    foreach ($users as $each_name) {
+      if ($_SESSION['brows_name'] == $each_name) {
+	$out .= '<td></td><td></td><td></td>';
+      } else {
+	$out .= '<td class=users><img src=dot.png width=100% height=2></td><td class=users><img src=dot.png width=100% height=2></td><td class=users><img src=dot.png width=100% height=2></td>';
+      }
+    }
+
+    $out .= '<td class=users><img src=dot.png width=100% height=2></td><tr></table>';
     return $out;
   }
 
@@ -751,12 +775,12 @@ class LangUpload {
   function handleAdminRequest() {
 
     // 桼ɲ
-    if (array_key_exists('add_user_submit', $_POST)) {
+    if (isset($_POST['add_user_submit'])) {
       $name = $_POST['add_user'];
 
       // ̾ιʸå
       if (preg_match('/[^A-Za-z0-9_]/', $name)) {
-	// !!! [A-Za-z0-9_] ˤơȤɽԤ
+	// !!! [A-Za-z0-9_] ˤơȤɽ桼󼨤
 	return;
       }
       // ǥ쥯ȥκ
@@ -767,9 +791,9 @@ class LangUpload {
       @chmod('data/'. $name, 0777);
 
       // users_file ˥桼ɲ
-      // !!! åޤ
       $fp = fopen($_SESSION['users_file'], 'a');
       if ($fp != false) {
+	flock($fp, LOCK_EX);
 	$line = $name. "\tdummy_password\n";
 	fwrite($fp, $line);
 	fclose($fp);
@@ -777,8 +801,8 @@ class LangUpload {
     }
 
     // ѥѹ
-    if (array_key_exists('change_password_submit', $_POST)) {
-      // !!! ɤǡ桼ʬΥѥɤΤѹơ񤭹
+    if (isset($_POST['change_password_submit'])) {
+      // ɤǡ桼ʬΥѥɤΤѹơ񤭹
       if (! isset($_POST['change_user'])) {
 	print 'no change_user<br>';
 	return;
@@ -788,6 +812,7 @@ class LangUpload {
       $replace_file = '';
       $fp = fopen($_SESSION['users_file'], 'r');
       if ($fp != false) {
+	flock($fp, LOCK_SH);
 	while (! feof($fp)) {
 	  $line = fgets($fp);
 	  $name = strtok($line, "\n\t ");
@@ -800,13 +825,14 @@ class LangUpload {
       }
       $fp = fopen($_SESSION['users_file'], 'w');
       if ($fp != false) {
+	flock($fp, LOCK_EX);
 	fwrite($fp, $replace_file);
 	fclose($fp);
       }
     }
 
     // 桼
-    if (array_key_exists('del_user_submit', $_POST)) {
+    if (isset($_POST['del_user_submit'])) {
     }
   }
 
@@ -814,18 +840,17 @@ class LangUpload {
   // ɽ
   function createCommentView() {
     $out = '';
-    $out .= '<table border=1 width=500>';
+    $out .= '<table border=1 width=80%><tr><th align=left>&nbsp;&nbsp;Comment</th></tr>';
 
     // !!! Ͽץ¸νȽʣʤΤޤȤ
-    $key = $this->getChapter(). '-'. $this->getNo();
-    $rev = array_key_exists($key, $this->comments_exist) ?
-      $this->comments_exist[$key] : 0;
-    $file = 'data/'. $_SESSION['brows_name']. '/c'. $key. '-'. $rev. '.txt';
+    $file = 'data/'. $_SESSION['brows_name']. '/c'.
+      $this->key. '-'. $this->rev. '.txt';
 
     $fp = @fopen($file, 'r');
     if ($fp == false) {
       $out .= '<tr><td>No Comments.</td></tr>';
     } else {
+      flock($fp, LOCK_SH);
       $header = true;
       while (! feof($fp)) {
 	$line = fgets($fp);
@@ -840,17 +865,20 @@ class LangUpload {
 
 	} else if ($header == true) {
 	  // ȥåΥإå
-	  $out .= '<tr><td>';
+	  $out .= '<tr><td class=cmt>';
 
 	  $header = false;
 	  $name = strtok($line, "\t ");
 	  $mtime = intval(strtok("\t "));
-	  $out .= date('Y-m-d', $mtime).
-	    '&nbsp; <strong>'. $name. '</strong><hr>';
+	  $out .= '<h3>&nbsp;'. date('Y-m-d H:i', $mtime).
+	    '&nbsp; '. $name. '</h3>';
 
 	} else {
 	  // Ȥɽ
-	  $out .= htmlspecialchars($line). '<br>';
+	  //$code = mb_detect_encoding($line, "auto", true);
+	  //$converted = mb_convert_encoding($line, 'EUC-JP', $code);
+	  $converted = $line;
+	  $out .= htmlspecialchars($converted). '';
 	}
       }
       fclose($fp);
@@ -864,7 +892,11 @@ class LangUpload {
   function createCommentForm() {
     $out = '';
 
-    // !!! ӥ󤬺ǿǤʤС
+    if ($this->rev_max != $this->rev) {
+      // ӥ󤬺ǿǤʤС
+      $out .= 'There is not comment form.<br>Please comment on the latest program.<br>';
+      return $out;
+    }
 
     $out .= '<form method=POST>';
 
@@ -880,19 +912,19 @@ class LangUpload {
   function handleCommentRequest($user) {
 
     // 桼ɲ
-    if (array_key_exists('comment_msg', $_POST)) {
+    if (isset($_POST['comment_msg'])) {
       // ץΥӥƱˤ
-      $this->createFilesHash($_SESSION['brows_name']);
-      $key = $this->getChapter(). '-'. $this->getNo();
-      $rev = array_key_exists($key, $this->programs_exist) ?
-	$this->programs_exist[$key] : 0;
-      $file = 'data/'. $_SESSION['brows_name']. '/c'. $key. '-'. $rev. '.txt';
+      $file = 'data/'. $_SESSION['brows_name']. '/c'.
+	$this->key. '-'. $this->rev. '.txt';
       $fp = fopen($file, 'a');
       if ($fp != false) {
 	$comment_data = $_SESSION['brows_name']. "\t". time(). "\n";
 
-	// !!! \n  \n&nbsp; ִ
-	$comment_data .= $_POST['comment_msg']. "\n\n";
+	// \n  \n&nbsp; ִ
+	$out .= preg_replace("/\n/", '<br>', $replaced);
+	$comment_data .=
+	  preg_replace("/\n/", "\n ", $_POST['comment_msg']). "\n\n";
+	flock($fp, LOCK_EX);
 	fwrite($fp, $comment_data);
 	fclose($fp);
 	$this->send_comment = true;
@@ -925,5 +957,195 @@ class LangUpload {
     }
     return $out;
   }
+
+
+  // ؤΥ󥯤
+  function createPrevNoLink() {
+    $out = '';
+    $prev_no = $this->getNo() - 1;
+    if ($prev_no <= 0) {
+      $out .= '<s>&lt;&lt;</s>';
+    } else {
+      $out .= '<a href=?no='. $prev_no. '>&lt;&lt</a>';
+    }
+    return $out;
+  }
+
+
+  // ؤΥ󥯤
+  function createNextNoLink() {
+    $out = '';
+    $next_no = $this->getNo() + 1;
+    if ($next_no > $_SESSION['problems_num'][$this->getChapter() -1]) {
+      $out .= '<s>&gt;&gt;</s>';
+    } else {
+      $out .= '<a href=?no='. $next_no. '>&gt;&gt</a>';
+    }
+    return $out;
+  }
+
+
+  // ӥΥ󥯤
+  function createRevLink() {
+
+    $out = '';
+
+    if ($this->rev_max == 0) {
+      return $out;
+    }
+    $out .= '<table border=1><tr><th>&nbsp;'. $this->key. '&nbsp;';
+
+    $first_index = (isset($this->comments_exist[$this->key])) ? 0 : 1;
+    if ($first_index != $this->rev_max) {
+      $out .= 'Rev';
+    }
+    $out .= '</th>';
+
+    if ($first_index != $this->rev_max) {
+      // rev:0 ΥȤСǥå 0 鳫Ϥ
+      for ($i = $first_index; $i <= $this->rev_max; ++$i) {
+	$out .= '<td><a href=?rev='. $i. '>&nbsp;'. $i. '</a></td>';
+      }
+    }
+    $out .= '</tr></table>';
+    return $out;
+  }
+
+
+  // ɻ֡ե빹μ
+  function selectUnreadPage() {
+    if (! isset($_SESSION['brows_name'])) {
+      // 桼λ꤬ɬפʤڡǤϡʤ
+      return;
+    }
+
+    // ɻ֤ɤ߽Ф
+    $login_name = $_SESSION['login_name'];
+    $brows_name = $_SESSION['brows_name'];
+    $dir_name = 'data/'. $login_name;
+
+    // ɻ: data/. $login_name. '/'. $brows_name. '-read.txt'
+    $file = $dir_name. '/'. $brows_name. '-read.txt';
+
+    $fp = @fopen($file, 'r');
+    if ($fp != false) {
+      flock($fp, LOCK_SH);
+      while (! feof($fp)) {
+	// : $key. "\t". $mtime
+	$line = fgets($fp);
+	$key = strtok($line, "\t ");
+	$mtime = intval(strtok("\t "));
+	if ($mtime != 0) {
+	  $this->read_mtime[$key] = $mtime;
+	}
+      }
+      fclose($fp);
+    }
+
+    // ե빹μ
+    $dir_name = 'data/'. $brows_name;
+    @$dp = opendir($dir_name);
+    if ($dp == false) {
+      return;
+    }
+    clearstatcache();
+    while (($file = readdir($dp)) !== false) {
+      // [pc]X-X-X.txt Ȥե
+      if (preg_match('/[pc]([\d]+-[\d]+-[\d]+)\.txt/', $file, $matches)) {
+
+	$key = $matches[1];
+	$mtime = filemtime($dir_name. '/'. $file);
+
+	if (! isset($this->files_mtime[$key])) {
+	  $this->files_mtime[$key] = $mtime;
+	} else {
+	  $time = $this->files_mtime[$key];
+	  if ($time < $mtime) {
+	    $this->files_mtime[$key] = $mtime;
+	  }
+	}
+      }
+    }
+    closedir($dp);
+
+    // ߱Υڡɰˤ
+    if (isset($_SESSION['no'])) {
+      $now_chapter = $this->getChapter();
+      $now_no = $this->getNo();
+      $now_rev = $this->rev;
+      $now_key = $now_chapter. '-'. $now_no. '-'. $now_rev;
+    } else {
+      $now_key = 'dummy_key';
+    }
+
+    // եι郎ɤ⿷Τʤ
+    foreach ($this->files_mtime as $key => $value) {
+      if ($key == $now_key) {
+	continue;
+      }
+      if (! isset($this->read_mtime[$key])) {
+	array_push($this->unread_pages, $key);
+      } else {
+	if ($this->read_mtime[$key] < $value) {
+	  array_push($this->unread_pages, $key);
+	}
+      }
+    }
+    sort($this->unread_pages);
+  }
+
+
+  /*!
+    \brief ̤ɥ󥯤ɤ
+  */
+  function createUnreadLink($message = 'Jump to unread page.') {
+    $out = '';
+
+    if (count($this->unread_pages) <= 0) {
+      // ̤ɥڡʤ
+      return $out;
+    }
+
+    // ̤ɥڡؤΥ󥯤
+    $key = array_shift($this->unread_pages);
+    if (preg_match('/([\d]+)-([\d]+)-([\d]+)/', $key, $matches)) {
+      $chapter = intval($matches[1]);
+      $no = intval($matches[2]);
+      $rev = intval($matches[3]);
+      $out .= '<a href=programView.php?chapter='. $chapter. '&no='. $no. '&rev='. $rev .'>'. $message. '</a>';
+    }
+    return $out;
+  }
+
+
+  // ɻ֤ι
+  // ɺѤˡ߱Ƥڡʬ򹹿Ƥ顢񤭹
+  function updateReadTime() {
+
+    // 
+    $now = time() + 1;
+    $now_chapter = $this->getChapter();
+    $now_no = $this->getNo();
+    $now_rev = $this->rev;
+    $key = $now_chapter. '-'. $now_no. '-'. $now_rev;
+    $this->read_mtime[$key] = $now;
+
+    // 񤭹
+    $login_name = $_SESSION['login_name'];
+    $brows_name = $_SESSION['brows_name'];
+    $dir_name = 'data/'. $login_name;
+    $file = $dir_name. '/'. $brows_name. '-read.txt';
+
+    $fp = @fopen($file, 'w');
+    if ($fp != false) {
+      flock($fp, LOCK_EX);
+      foreach ($this->read_mtime as $key => $value) {
+	// : $key. "\t". $mtime
+	$line = $key. "\t". $value. "\n";
+	fwrite($fp, $line);
+      }
+      fclose($fp);
+    }
+  }
 };
 ?>
