@@ -15,13 +15,14 @@ require_once("includes/langUpload.php");
 $user = new UserAuthorize();
 
 if ($user->requireReload() || ! $user->isLogin()) {
-  // !!! chapter, no ̵ʾˤ褦ˤ
   header('Location: index.php');
   return;
 }
 
 $lang = new LangUpload();
 $lang->handleCommentRequest($user);
+$lang->updateReadTime();
+
 if ($lang->sendComment()) {
   header('Location: programView.php#comment');
   return;
